package de.hhhammer.dchat.bot;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.zaxxer.hikari.HikariConfig;
import com.zaxxer.hikari.HikariDataSource;
import de.hhhammer.dchat.bot.openai.ChatGPTService;
import de.hhhammer.dchat.db.PostgresServerDBService;
import de.hhhammer.dchat.db.PostgresUserDBService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.http.HttpClient;

public final class App {
    private static final Logger logger = LoggerFactory.getLogger(App.class);

    public static void main(final String[] args) {
        final String discordApiKey = System.getenv("DISCORD_API_KEY");
        if (discordApiKey == null) {
            logger.error("Missing environment variables: DISCORD_API_KEY");
            System.exit(1);
        }
        final String openaiApiKey = System.getenv("OPENAI_API_KEY");
        if (openaiApiKey == null) {
            logger.error("Missing environment variables: OPENAI_API_KEY");
            System.exit(1);
        }
        final String postgresUser = System.getenv("POSTGRES_USER");
        final String postgresPassword = System.getenv("POSTGRES_PASSWORD");
        final String postgresUrl = System.getenv("POSTGRES_URL");
        if (postgresUser == null || postgresPassword == null || postgresUrl == null) {
            logger.error("Missing environment variables: POSTGRES_USER and/or POSTGRES_PASSWORD and/or POSTGRES_URL");
            System.exit(1);
        }

        final var chatGPTService = new ChatGPTService(openaiApiKey, HttpClient.newHttpClient(), new ObjectMapper());

        final var config = new HikariConfig();
        config.setJdbcUrl(postgresUrl);
        config.setUsername(postgresUser);
        config.setPassword(postgresPassword);

        try (var ds = new HikariDataSource(config)) {
            final var serverDBService = new PostgresServerDBService(ds);
            final var userDBService = new PostgresUserDBService(ds);

            final var discordBot = new DiscordBot(serverDBService, userDBService, chatGPTService, discordApiKey);
            discordBot.run();
        }
    }
}
