package de.hhhammer.dchat.bot;

import de.hhhammer.dchat.bot.discord.MessageCreateHandler;
import de.hhhammer.dchat.bot.discord.ServerMessageHandler;
import de.hhhammer.dchat.bot.discord.UserMessageHandler;
import de.hhhammer.dchat.bot.openai.ChatGPTService;
import de.hhhammer.dchat.db.ServerDBService;
import de.hhhammer.dchat.db.UserDBService;
import org.javacord.api.DiscordApi;
import org.javacord.api.DiscordApiBuilder;
import org.javacord.api.interaction.SlashCommand;
import org.javacord.api.interaction.SlashCommandInteraction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

public final class DiscordBot implements Runnable {
    private static final Logger logger = LoggerFactory.getLogger(DiscordBot.class);

    private final ServerDBService serverDBService;
    private final UserDBService userDBService;
    private final ChatGPTService chatGPTService;
    private final String discordApiKey;

    public DiscordBot(final ServerDBService serverDBService, final UserDBService userDBService, final ChatGPTService chatGPTService, final String discordApiKey) {
        this.serverDBService = serverDBService;
        this.userDBService = userDBService;
        this.chatGPTService = chatGPTService;
        this.discordApiKey = discordApiKey;
    }

    @Override
    public void run() {
        logger.info("Starting Discord application");
        final DiscordApi discordApi = new DiscordApiBuilder()
                .setToken(discordApiKey)
                .login()
                .join();
        discordApi.setMessageCacheSize(10, 60 * 60);
        final var future = new CompletableFuture<Void>();
        Runtime.getRuntime().addShutdownHook(Thread.ofVirtual().unstarted(() -> {
            logger.info("Shutting down Discord application");
            discordApi.disconnect().thenAccept(future::complete);
        }));
        final SlashCommand token = SlashCommand.with("tokens", "Check how many tokens where spend on this server")
                .createGlobal(discordApi)
                .join();

        discordApi.addSlashCommandCreateListener(event -> {
            logger.debug("Event? " + event.getSlashCommandInteraction().getFullCommandName());
            final SlashCommandInteraction command = event.getSlashCommandInteraction();
            if (token.getFullCommandNames().contains(command.getFullCommandName())) {
                event.getInteraction()
                        .respondLater()
                        .orTimeout(30, TimeUnit.SECONDS)
                        .thenAccept((interactionOriginalResponseUpdater) -> {
                            final long tokens = event.getInteraction().getServer().isPresent() ?
                                    this.serverDBService.tokensOfLast30Days(String.valueOf(event.getInteraction().getServer().get().getId())) :
                                    this.userDBService.tokensOfLast30Days(String.valueOf(event.getInteraction().getUser().getId()));
                            interactionOriginalResponseUpdater.setContent("" + tokens).update();
                        });
                return;
            }
        });


        discordApi.addMessageCreateListener(new MessageCreateHandler(new ServerMessageHandler(serverDBService, chatGPTService)));
        discordApi.addMessageCreateListener(new MessageCreateHandler(new UserMessageHandler(userDBService, chatGPTService)));

        // Print the invite url of your bot
        logger.info("You can invite the bot by using the following url: " + discordApi.createBotInvite());

        try {
            future.get();
        } catch (InterruptedException | ExecutionException e) {
            throw new RuntimeException(e);
        }
    }
}
