package de.hhhammer.dchat.bot.discord;

import de.hhhammer.dchat.bot.openai.ChatGPTRequestBuilder;
import de.hhhammer.dchat.bot.openai.ChatGPTService;
import de.hhhammer.dchat.bot.openai.MessageContext.ReplyInteraction;
import de.hhhammer.dchat.bot.openai.ResponseException;
import de.hhhammer.dchat.bot.openai.models.ChatGPTRequest;
import de.hhhammer.dchat.bot.openai.models.ChatGPTResponse;
import de.hhhammer.dchat.db.ServerDBService;
import de.hhhammer.dchat.db.models.server.ServerConfig;
import de.hhhammer.dchat.db.models.server.ServerMessage;
import org.javacord.api.entity.DiscordEntity;
import org.javacord.api.entity.message.Message;
import org.javacord.api.entity.message.MessageReference;
import org.javacord.api.entity.message.MessageType;
import org.javacord.api.event.message.MessageCreateEvent;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.List;
import java.util.Optional;

public final class ServerMessageHandler implements MessageHandler {

    private static final Logger logger = LoggerFactory.getLogger(ServerMessageHandler.class);
    private final ServerDBService serverDBService;
    private final ChatGPTService chatGPTService;

    public ServerMessageHandler(final ServerDBService serverDBService, final ChatGPTService chatGPTService) {
        this.serverDBService = serverDBService;
        this.chatGPTService = chatGPTService;
    }

    @Override
    public void handle(final MessageCreateEvent event) throws ResponseException, IOException, InterruptedException {
        final String content = extractContent(event);
        final long serverId = event.getServer().get().getId();
        final String systemMessage = this.serverDBService.getConfig(String.valueOf(serverId)).get().systemMessage();
        final List<ReplyInteraction> messageContext = event.getMessage().getType() == MessageType.REPLY ? getContextMessages(event) : List.of();
        final ChatGPTRequest request = new ChatGPTRequestBuilder().contextRequest(messageContext, content, systemMessage);
        final ChatGPTResponse response = this.chatGPTService.submit(request);
        if (response.choices().isEmpty()) {
            event.getMessage().reply("No response available");
            return;
        }
        final String answer = response.choices().get(0).message().content();
        logServerMessage(event, response.usage().totalTokens());
        event.getMessage().reply(answer);
    }

    @Override
    public boolean isAllowed(final MessageCreateEvent event) {
        if (event.getServer().isEmpty()) {
            return false;
        }

        final long serverId = event.getServer().get().getId();
        final Optional<ServerConfig> config = this.serverDBService.getConfig(String.valueOf(serverId));
        if (config.isEmpty()) {
            logger.debug("Not allowed with id: " + serverId);
            return false;
        }
        return true;
    }

    @Override
    public boolean exceedsRate(final MessageCreateEvent event) {
        final String serverId = String.valueOf(event.getServer().get().getId());
        final Optional<ServerConfig> config = this.serverDBService.getConfig(serverId);
        if (config.isEmpty()) {
            logger.error("Missing configuration for server with id: " + serverId);
            return true;
        }
        final int rateLimit = config.get().rateLimit();
        final int countMessagesInLastMinute = this.serverDBService.countMessagesInLastMinute(serverId);

        return countMessagesInLastMinute >= rateLimit;
    }

    @Override
    public boolean canHandle(final MessageCreateEvent event) {
        return event.isServerMessage();
    }

    private void logServerMessage(final MessageCreateEvent event, final int tokens) {
        final long serverId = event.getServer().map(DiscordEntity::getId).get();
        final long userId = event.getMessageAuthor().getId();

        final var serverMessage = new ServerMessage.NewServerMessage(String.valueOf(serverId), userId, tokens);
        this.serverDBService.addMessage(serverMessage);
    }

    private String extractContent(final MessageCreateEvent event) {
        final long ownId = event.getApi().getYourself().getId();
        return event.getMessageContent().replaceFirst("<" + ownId + "> ", "");
    }

    @NotNull
    private List<ReplyInteraction> getContextMessages(final MessageCreateEvent event) {
        return event.getMessage()
                .getMessageReference()
                .map(MessageReference::getMessage)
                .flatMap(m -> m)
                .map(Message::getReadableContent)
                .map(ReplyInteraction::new)
                .stream()
                .toList();
    }
}
