package de.hhhammer.dchat.bot.discord;

import de.hhhammer.dchat.bot.openai.ChatGPTRequestBuilder;
import de.hhhammer.dchat.bot.openai.ChatGPTService;
import de.hhhammer.dchat.bot.openai.MessageContext.PreviousInteraction;
import de.hhhammer.dchat.bot.openai.ResponseException;
import de.hhhammer.dchat.bot.openai.models.ChatGPTRequest;
import de.hhhammer.dchat.bot.openai.models.ChatGPTResponse;
import de.hhhammer.dchat.db.UserDBService;
import de.hhhammer.dchat.db.models.user.UserConfig;
import de.hhhammer.dchat.db.models.user.UserMessage;
import org.javacord.api.event.message.MessageCreateEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.List;
import java.util.Optional;

public final class UserMessageHandler implements MessageHandler {
    private static final Logger logger = LoggerFactory.getLogger(UserMessageHandler.class);
    private final UserDBService userDBService;
    private final ChatGPTService chatGPTService;

    public UserMessageHandler(final UserDBService userDBService, final ChatGPTService chatGPTService) {
        this.userDBService = userDBService;
        this.chatGPTService = chatGPTService;
    }

    @Override
    public void handle(final MessageCreateEvent event) throws ResponseException, IOException, InterruptedException {
        final String content = event.getReadableMessageContent();
        final String userId = String.valueOf(event.getMessageAuthor().getId());
        final UserConfig config = this.userDBService.getConfig(userId).get();
        final String systemMessage = config.systemMessage();
        final List<PreviousInteraction> context = this.userDBService.getLastMessages(userId, config.contextLength())
                .stream()
                .map(userMessage -> new PreviousInteraction(userMessage.question(), userMessage.answer()))
                .toList();
        final ChatGPTRequest request = new ChatGPTRequestBuilder().contextRequest(context, content, systemMessage);
        final ChatGPTResponse response = this.chatGPTService.submit(request);
        if (response.choices().isEmpty()) {
            event.getMessage().reply("No response available");
            return;
        }
        final String answer = response.choices().get(0).message().content();
        logUserMessage(event, content, answer, response.usage().totalTokens());
        event.getChannel().sendMessage(answer);
    }

    @Override
    public boolean isAllowed(final MessageCreateEvent event) {
        if (event.getServer().isPresent()) {
            return false;
        }

        final long userId = event.getMessageAuthor().getId();
        final Optional<UserConfig> config = this.userDBService.getConfig(String.valueOf(userId));
        if (config.isEmpty()) {
            logger.debug("Not allowed with id: " + userId);
            return false;
        }
        return true;
    }

    @Override
    public boolean exceedsRate(final MessageCreateEvent event) {
        final String userId = String.valueOf(event.getMessageAuthor().getId());
        final Optional<UserConfig> config = this.userDBService.getConfig(userId);
        if (config.isEmpty()) {
            logger.error("Missing configuration for userId with id: " + userId);
            return true;
        }
        final int rateLimit = config.get().rateLimit();
        final int countMessagesInLastMinute = this.userDBService.countMessagesInLastMinute(userId);

        return countMessagesInLastMinute >= rateLimit;
    }

    @Override
    public boolean canHandle(final MessageCreateEvent event) {
        return event.isPrivateMessage();
    }

    private void logUserMessage(final MessageCreateEvent event, final String question, final String answer, final int tokens) {
        final long userId = event.getMessageAuthor().getId();

        final UserMessage.NewUserMessage userMessage = new UserMessage.NewUserMessage(String.valueOf(userId), question, answer, tokens);
        this.userDBService.addMessage(userMessage);
    }
}
