package de.hhhammer.dchat.db;

import de.hhhammer.dchat.db.models.server.ServerConfig;
import de.hhhammer.dchat.db.models.server.ServerMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.sql.DataSource;
import java.sql.*;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.Optional;
import java.util.stream.StreamSupport;

public final class PostgresServerDBService implements ServerDBService {
    private static final Logger logger = LoggerFactory.getLogger(PostgresServerDBService.class);
    private final DataSource dataSource;

    public PostgresServerDBService(final DataSource dataSource) {
        this.dataSource = dataSource;
    }

    @Override
    public Optional<ServerConfig> getConfig(final String serverId) {
        final var getServerConfig = """
                SELECT * FROM server_configs WHERE server_id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, serverId);
            final ResultSet resultSet = pstmt.executeQuery();
            final Iterable<ServerConfig> iterable = () -> new ResultSetIterator<>(resultSet, new ServerConfig.ServerConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).findFirst();
        } catch (final SQLException e) {
            logger.error("Getting configuration for server with id: " + serverId, e);
        } catch (final ResultSetIteratorException e) {
            logger.error("Iterating over ServerConfig ResultSet for server with id: " + serverId, e);
        }
        return Optional.empty();
    }

    @Override
    public List<ServerConfig> getAllConfigs() throws DBException {
        final var getAllowedServerSql = """
                SELECT * FROM server_configs
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getAllowedServerSql)
        ) {
            final ResultSet resultSet = pstmt.executeQuery();
            final Iterable<ServerConfig> iterable = () -> new ResultSetIterator<>(resultSet, new ServerConfig.ServerConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).toList();
        } catch (final SQLException e) {
            throw new DBException("Loading all configs", e);
        } catch (final ResultSetIteratorException e) {
            throw new DBException("Iterating over configs", e);
        }
    }

    @Override
    public Optional<ServerConfig> getConfigBy(final long id) throws DBException {
        final var getServerConfig = """
                SELECT * FROM server_configs WHERE id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setLong(1, id);
            final ResultSet resultSet = pstmt.executeQuery();
            final Iterable<ServerConfig> iterable = () -> new ResultSetIterator<>(resultSet, new ServerConfig.ServerConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).findFirst();
        } catch (SQLException e) {
            throw new DBException("Getting configuration with id: " + id, e);
        } catch (ResultSetIteratorException e) {
            throw new DBException("Iterating over ServerConfig ResultSet for id: " + id, e);
        }
    }

    @Override
    public void addConfig(final ServerConfig.NewServerConfig newServerConfig) throws DBException {
        final var getServerConfig = """
                INSERT INTO server_configs (server_id, system_message, rate_limit) VALUES (?,?,?)
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newServerConfig.serverId());
            pstmt.setString(2, newServerConfig.systemMessage());
            pstmt.setInt(3, newServerConfig.rateLimit());
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config added for server with id: " + newServerConfig.serverId());
            }
        } catch (final SQLException e) {
            throw new DBException("Adding configuration to server with id: " + newServerConfig.serverId(), e);
        }
    }

    @Override
    public void updateConfig(final long id, final ServerConfig.NewServerConfig newServerConfig) throws DBException {
        final var getServerConfig = """
                UPDATE server_configs SET system_message = ?, rate_limit = ?, server_id = ? WHERE id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newServerConfig.systemMessage());
            pstmt.setInt(2, newServerConfig.rateLimit());
            pstmt.setString(3, newServerConfig.serverId());
            pstmt.setLong(4, id);
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config update for server with id: " + newServerConfig.serverId());
            }
        } catch (final SQLException e) {
            throw new DBException("Adding configuration to server with id: " + newServerConfig.serverId(), e);
        }
    }

    @Override
    public void deleteConfig(final long id) throws DBException {
        final var getServerConfig = """
                DELETE FROM server_configs WHERE id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setLong(1, id);
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config deleted for server with id: " + id);
            }
        } catch (final SQLException e) {
            throw new DBException("Deleting configuration for server with id: " + id, e);
        }
    }

    @Override
    public int countMessagesInLastMinute(final String serverId) {
        final var getServerConfig = """
                SELECT count(*) FROM server_messages WHERE server_id = ? AND time <= ? and time >= ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, serverId);
            final var now = Instant.now();
            pstmt.setTimestamp(2, Timestamp.from(now));
            pstmt.setTimestamp(3, Timestamp.from(now.minus(1, ChronoUnit.MINUTES)));
            final ResultSet resultSet = pstmt.executeQuery();
            if (resultSet.next()) return resultSet.getInt(1);
        } catch (final SQLException e) {
            logger.error("Getting messages for server with id: " + serverId, e);
        } catch (final ResultSetIteratorException e) {
            logger.error("Iterating over ServerMessages ResultSet for server with id: " + serverId, e);
        }
        return Integer.MAX_VALUE;
    }

    @Override
    public void addMessage(final ServerMessage.NewServerMessage serverMessage) {
        final var getServerConfig = """
                INSERT INTO server_messages (server_id, user_id, tokens) VALUES (?,?,?)
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, serverMessage.serverId());
            pstmt.setLong(2, serverMessage.userId());
            pstmt.setInt(3, serverMessage.tokens());
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No message added for server with id: " + serverMessage.serverId());
            }
        } catch (final SQLException e) {
            logger.error("Adding message to server with id: " + serverMessage.serverId(), e);
        }
    }

    @Override
    public long tokensOfLast30Days(final String serverId) {
        final var countTokensOfLast30Days = """
                SELECT sum(tokens) FROM server_messages WHERE server_id = ? AND time < ? AND time >= ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(countTokensOfLast30Days)
        ) {
            pstmt.setString(1, serverId);
            final var now = Instant.now();
            pstmt.setTimestamp(2, Timestamp.from(now));
            pstmt.setTimestamp(3, Timestamp.from(now.minus(30, ChronoUnit.DAYS)));
            final ResultSet resultSet = pstmt.executeQuery();
            if (resultSet.next()) return resultSet.getLong(1);
        } catch (final SQLException e) {
            logger.error("Counting tokens of the last 30 days from server with id: " + serverId, e);
        }
        logger.error("No tokens found for server with id: " + serverId);
        return 0;
    }
}
