package de.hhhammer.dchat.db;

import de.hhhammer.dchat.db.models.user.UserConfig;
import de.hhhammer.dchat.db.models.user.UserMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.sql.DataSource;
import java.sql.*;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.Optional;
import java.util.stream.StreamSupport;

public final class PostgresUserDBService implements UserDBService {
    private static final Logger logger = LoggerFactory.getLogger(PostgresUserDBService.class);
    private final DataSource dataSource;

    public PostgresUserDBService(final DataSource dataSource) {
        this.dataSource = dataSource;
    }

    @Override
    public Optional<UserConfig> getConfig(final String userId) {
        final var getServerConfig = """
                SELECT * FROM user_configs WHERE user_id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, userId);
            final ResultSet resultSet = pstmt.executeQuery();
            final Iterable<UserConfig> iterable = () -> new ResultSetIterator<>(resultSet, new UserConfig.UserConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).findFirst();
        } catch (final SQLException e) {
            logger.error("Getting configuration for user with id: " + userId, e);
        } catch (final ResultSetIteratorException e) {
            logger.error("Iterating over ServerConfig ResultSet for user with id: " + userId, e);
        }
        return Optional.empty();
    }

    @Override
    public Optional<UserConfig> getConfigBy(final long id) throws DBException {
        final var getServerConfig = """
                SELECT * FROM user_configs WHERE id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setLong(1, id);
            final ResultSet resultSet = pstmt.executeQuery();
            final Iterable<UserConfig> iterable = () -> new ResultSetIterator<>(resultSet, new UserConfig.UserConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).findFirst();
        } catch (final SQLException e) {
            throw new DBException("Getting configuration id: " + id, e);
        } catch (final ResultSetIteratorException e) {
            throw new DBException("Iterating over UserConfig ResultSet with id: " + id, e);
        }
    }

    @Override
    public List<UserConfig> getAllConfigs() throws DBException {
        final var getServerConfig = """
                SELECT * FROM user_configs
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            final ResultSet resultSet = pstmt.executeQuery();
            final Iterable<UserConfig> iterable = () -> new ResultSetIterator<>(resultSet, new UserConfig.UserConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).toList();
        } catch (final SQLException e) {
            throw new DBException("Getting all configurations", e);
        } catch (final ResultSetIteratorException e) {
            throw new DBException("Iterating over all UserConfig ResultSet", e);
        }
    }

    @Override
    public void addConfig(UserConfig.NewUserConfig newUserConfig) throws DBException {
        final var getServerConfig = """
                INSERT INTO user_configs (user_id, system_message, context_length, rate_limit) VALUES (?,?,?,?)
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newUserConfig.userId());
            pstmt.setString(2, newUserConfig.systemMessage());
            pstmt.setInt(3, newUserConfig.contextLength());
            pstmt.setInt(4, newUserConfig.rateLimit());
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config added for user with id: " + newUserConfig.userId());
            }
        } catch (final SQLException e) {
            throw new DBException("Adding configuration for user with id: " + newUserConfig.userId(), e);
        }
    }

    @Override
    public void updateConfig(final long id, final UserConfig.NewUserConfig newUserConfig) throws DBException {
        final var getServerConfig = """
                UPDATE user_configs SET system_message = ?, context_length = ?, rate_limit = ?, user_id = ? WHERE id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newUserConfig.systemMessage());
            pstmt.setInt(2, newUserConfig.rateLimit());
            pstmt.setLong(3, newUserConfig.contextLength());
            pstmt.setString(4, newUserConfig.userId());
            pstmt.setLong(5, id);
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config update with id: " + id);
            }
        } catch (final SQLException e) {
            throw new DBException("Updating configuration with id: " + id, e);
        }
    }

    @Override
    public void deleteConfig(final long id) throws DBException {
        final var getServerConfig = """
                DELETE FROM user_configs WHERE id = ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setLong(1, id);
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config deleted for user with id: " + id);
            }
        } catch (final SQLException e) {
            throw new DBException("Deleting configuration with id: " + id, e);
        }
    }

    @Override
    public int countMessagesInLastMinute(final String userId) {
        final var getServerConfig = """
                SELECT count(*) FROM user_messages WHERE user_id = ? AND time <= ? and time >= ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, userId);
            final var now = Instant.now();
            pstmt.setTimestamp(2, Timestamp.from(now));
            pstmt.setTimestamp(3, Timestamp.from(now.minus(1, ChronoUnit.MINUTES)));
            final ResultSet resultSet = pstmt.executeQuery();
            if (resultSet.next()) return resultSet.getInt(1);
        } catch (final SQLException e) {
            logger.error("Getting messages for user with id: " + userId, e);
        } catch (final ResultSetIteratorException e) {
            logger.error("Iterating over ServerMessages ResultSet for user with id: " + userId, e);
        }
        return Integer.MAX_VALUE;
    }

    @Override
    public void addMessage(final UserMessage.NewUserMessage newUserMessage) {
        final var getServerConfig = """
                INSERT INTO user_messages (user_id, question, answer, tokens) VALUES (?,?,?,?)
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newUserMessage.userId());
            pstmt.setString(2, newUserMessage.question());
            pstmt.setString(3, newUserMessage.answer());
            pstmt.setInt(4, newUserMessage.tokens());
            final int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No message added for user with id: " + newUserMessage.userId());
            }
        } catch (final SQLException e) {
            logger.error("Adding message to user with id: " + newUserMessage.userId(), e);
        }
    }

    @Override
    public List<UserMessage> getLastMessages(final String userId, final int limit) {
        final var getLastMessages = """
                SELECT * FROM user_messages WHERE user_id = ? ORDER BY time DESC LIMIT ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(getLastMessages)
        ) {
            pstmt.setString(1, userId);
            pstmt.setInt(2, limit);
            final ResultSet resultSet = pstmt.executeQuery();
            final Iterable<UserMessage> iterable = () -> new ResultSetIterator<>(resultSet, new UserMessage.UserMessageResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).toList();
        } catch (final SQLException e) {
            logger.error("Fetching last messages for user whit id: " + userId, e);
        } catch (final ResultSetIteratorException e) {
            logger.error("Iterating over messages ResultSet from user with id: " + userId, e);
        }
        return List.of();
    }

    @Override
    public long tokensOfLast30Days(final String userId) {
        final var countTokensOfLast30Days = """
                SELECT sum(tokens) FROM user_messages WHERE user_id = ? AND time < ? AND time >= ?
                """;
        try (final Connection con = dataSource.getConnection();
             final PreparedStatement pstmt = con.prepareStatement(countTokensOfLast30Days)
        ) {
            pstmt.setString(1, userId);
            final var now = Instant.now();
            pstmt.setTimestamp(2, Timestamp.from(now));
            pstmt.setTimestamp(3, Timestamp.from(now.minus(30, ChronoUnit.DAYS)));
            final ResultSet resultSet = pstmt.executeQuery();
            if (resultSet.next()) return resultSet.getLong(1);
        } catch (final SQLException e) {
            logger.error("Counting tokens of the last 30 days from user with id: " + userId, e);
        }
        logger.error("No tokens found for user with id: " + userId);
        return 0;
    }
}
