package de.hhhammer.dchat.monolith;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.zaxxer.hikari.HikariConfig;
import com.zaxxer.hikari.HikariDataSource;
import de.hhhammer.dchat.bot.DiscordBot;
import de.hhhammer.dchat.bot.openai.ChatGPTService;
import de.hhhammer.dchat.db.PostgresServerDBService;
import de.hhhammer.dchat.db.PostgresUserDBService;
import de.hhhammer.dchat.web.AppConfig;
import de.hhhammer.dchat.web.WebAPI;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.http.HttpClient;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Executors;

public final class App {
    private static final Logger logger = LoggerFactory.getLogger(App.class);

    public static void main(final String[] args) {
        final String discordApiKey = System.getenv("DISCORD_API_KEY");
        if (discordApiKey == null) {
            logger.error("Missing environment variables: DISCORD_API_KEY");
            System.exit(1);
        }
        final String openaiApiKey = System.getenv("OPENAI_API_KEY");
        if (openaiApiKey == null) {
            logger.error("Missing environment variables: OPENAI_API_KEY");
            System.exit(1);
        }
        final String postgresUser = System.getenv("POSTGRES_USER");
        final String postgresPassword = System.getenv("POSTGRES_PASSWORD");
        final String postgresUrl = System.getenv("POSTGRES_URL");
        if (postgresUser == null || postgresPassword == null || postgresUrl == null) {
            logger.error("Missing environment variables: POSTGRES_USER and/or POSTGRES_PASSWORD and/or POSTGRES_URL");
            System.exit(1);
        }
        final String apiPortStr = System.getenv("API_PORT") != null ? System.getenv("API_PORT") : "8080";
        final int apiPort = Integer.parseInt(apiPortStr);
        final boolean debug = "true".equals(System.getenv("API_DEBUG"));

        final var chatGPTService = new ChatGPTService(openaiApiKey, HttpClient.newHttpClient(), new ObjectMapper());

        final var config = new HikariConfig();
        config.setJdbcUrl(postgresUrl);
        config.setUsername(postgresUser);
        config.setPassword(postgresPassword);

        try (final var ds = new HikariDataSource(config)) {
            final var serverDBService = new PostgresServerDBService(ds);
            final var userDBService = new PostgresUserDBService(ds);

            final var discordBot = new DiscordBot(serverDBService, userDBService, chatGPTService, discordApiKey);
            final var appConfig = new AppConfig(apiPort, debug);
            final var webApi = new WebAPI(serverDBService, userDBService, appConfig);
            run(discordBot, webApi);
        } catch (InterruptedException e) {
            logger.error("Application failed", e);
        }
    }

    private static void run(final Runnable... apps) throws InterruptedException {
        final List<Callable<Void>> callableApps = Arrays.stream(apps).sequential().map(runnable -> (Callable<Void>) () -> {
            runnable.run();
            return null;
        }).toList();
        try (final var executorService = Executors.newFixedThreadPool(apps.length)) {
            executorService.invokeAll(callableApps);
        }
    }
}
