package de.hhhammer.dchat.web.server;

import de.hhhammer.dchat.db.DBException;
import de.hhhammer.dchat.db.ServerDBService;
import de.hhhammer.dchat.db.models.server.ServerConfig;
import io.javalin.apibuilder.CrudHandler;
import io.javalin.http.Context;
import io.javalin.http.HttpStatus;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Optional;

public final class ConfigCrudHandler implements CrudHandler {
    private static final Logger logger = LoggerFactory.getLogger(ConfigCrudHandler.class);

    private final ServerDBService serverDBService;

    public ConfigCrudHandler(final ServerDBService serverDBService) {
        this.serverDBService = serverDBService;
    }

    @Override
    public void create(@NotNull final Context context) {
        final ServerConfig.NewServerConfig body = context.bodyAsClass(ServerConfig.NewServerConfig.class);
        try {
            this.serverDBService.addConfig(body);
        } catch (DBException e) {
            logger.error("Adding new server configuration", e);
            context.status(HttpStatus.INTERNAL_SERVER_ERROR);
        }
        context.status(HttpStatus.CREATED);
    }

    @Override
    public void delete(@NotNull final Context context, @NotNull final String s) {
        final var id = Long.parseLong(s);
        try {
            this.serverDBService.deleteConfig(id);
            context.status(HttpStatus.NO_CONTENT);
        } catch (DBException e) {
            logger.error("Deleting configuration with id: " + s, e);
            context.status(HttpStatus.INTERNAL_SERVER_ERROR);
        }
    }

    @Override
    public void getAll(@NotNull final Context context) {
        try {
            final List<ServerConfig> allowedServers = this.serverDBService.getAllConfigs();
            context.json(allowedServers);
        } catch (DBException e) {
            logger.error("Getting all server configs", e);
            context.status(HttpStatus.INTERNAL_SERVER_ERROR);
        }
    }

    @Override
    public void getOne(@NotNull final Context context, @NotNull final String s) {
        final var id = Long.parseLong(s);
        try {
            final Optional<ServerConfig> server = this.serverDBService.getConfigBy(id);
            if (server.isEmpty()) {
                context.status(HttpStatus.NOT_FOUND);
                return;
            }
            context.json(server.get());
        } catch (DBException e) {
            logger.error("Searching for config with id: " + s, e);
            context.status(HttpStatus.INTERNAL_SERVER_ERROR);
        }
    }

    @Override
    public void update(@NotNull final Context context, @NotNull final String idString) {
        final ServerConfig.NewServerConfig body = context.bodyAsClass(ServerConfig.NewServerConfig.class);
        final var id = Long.parseLong(idString);

        try {
            this.serverDBService.updateConfig(id, body);
        } catch (DBException e) {
            logger.error("Updating allowed server with id: " + idString, e);
            context.status(HttpStatus.INTERNAL_SERVER_ERROR);
        }
    }
}
